// Prevent Visual Studio Intellisense from defining _WIN32 and _MSC_VER when we use 
// Visual Studio to edit Linux or Borland C++ code.
#ifdef __linux__
#	undef _WIN32
#endif // __linux__
#if defined(__GNUC__) || defined(__BORLANDC__)
#	undef _MSC_VER
#endif // defined(__GNUC__) || defined(__BORLANDC__)

#include "Globals.h"

void CleanUp(void)
{
	if (!disablecli2) releasetcpcli(s2);
	releasetcpcli(s1);
	if (enablespy) fclose(fspy);
}

int handlecli(SOCKET sockcli, void* pParam)
{
	char databuf[MAX_BUF_LEN];
	int nbBytesReceived = 0;
	char spyfilename[MAX_BUF_LEN];
	int i = 0;

	if (enablespy) 
	{
		sprintf(spyfilename, "spy_%s.txt", strtime_fns());

		fspy = fopen(spyfilename, "w");
		if (fspy == NULL)
		{
			printf("Unable to create spy file.\n");
			return EXIT_FAILURE;
		}
	}

	if (inittcpcli(&s1, cli1address, cli1port) != EXIT_SUCCESS)
	{
		if (enablespy) fclose(fspy);
		return EXIT_FAILURE;
	}
	if (!disablecli2)
	{
		if (inittcpcli(&s2, cli2address, cli2port) != EXIT_SUCCESS)
		{
			releasetcpcli(s1);
			if (enablespy) fclose(fspy);
			return EXIT_FAILURE;
		}
	}

	for (;;)
	{
		fd_set sock_set;
		int iResult = SOCKET_ERROR;
		struct timeval tv;

		tv.tv_sec = (long)(timeout/1000);
		tv.tv_usec = (long)((timeout%1000)*1000);

		// Initialize a fd_set and add the socket to it.
		FD_ZERO(&sock_set); 
		FD_SET(sockcli, &sock_set);

		iResult = select((int)sockcli+1, &sock_set, NULL, NULL, &tv);

		// Remove the socket from the set.
		// No need to use FD_ISSET() here, as we only have one socket the return value of select() is 
		// sufficient to know what happened.
		FD_CLR(sockcli, &sock_set); 

		switch (iResult)
		{
		case SOCKET_ERROR:
			CleanUp();
			return EXIT_FAILURE;
		case 0:
			// The timeout on select() occured.
			break;
		default:
			memset(databuf, 0, sizeof(databuf));
			nbBytesReceived = recv(sockcli, databuf, sizeof(databuf), 0);
			if (nbBytesReceived <= 0)
			{
				CleanUp();
				return EXIT_FAILURE;
			}

			if (enablespy) 
			{
				if (charmode) 
				{
					for (i = 0; i < nbBytesReceived; i++)
					{
						if (fputc((int)databuf[i], fspy) == EOF)
						{
							printf("fputc() failed.\n");
						}
					}
				}
				else
				{
					for (i = 0; i < nbBytesReceived; i++)
					{
						if (fprintf(fspy, "%02x ", (unsigned int)((unsigned char)databuf[i])) <= 0)
						{
							printf("fprintf() failed.\n");
						}
					}
				}
				if (fflush(fspy) != EXIT_SUCCESS)
				{
					printf("fflush() failed.\n");
				}
			}

			if (sendall(s1, databuf, nbBytesReceived) != EXIT_SUCCESS)
			{
				CleanUp();
				return EXIT_FAILURE;
			}
			if (!disablecli2) 
			{
				if (sendall(s2, databuf, nbBytesReceived) != EXIT_SUCCESS)
				{
					CleanUp();
					return EXIT_FAILURE;
				}
			}
			break;
		}

		if (!readonly)
		{
			tv.tv_sec = (long)(timeout/1000);
			tv.tv_usec = (long)((timeout%1000)*1000);

			// Initialize a fd_set and add the socket to it.
			FD_ZERO(&sock_set); 
			FD_SET(s1, &sock_set);

			iResult = select((int)s1+1, &sock_set, NULL, NULL, &tv);

			// Remove the socket from the set.
			// No need to use FD_ISSET() here, as we only have one socket the return value of select() is 
			// sufficient to know what happened.
			FD_CLR(s1, &sock_set); 

			switch (iResult)
			{
			case SOCKET_ERROR:
				CleanUp();
				return EXIT_FAILURE;
			case 0:
				// The timeout on select() occured.
				break;
			default:
				memset(databuf, 0, sizeof(databuf));
				nbBytesReceived = recv(s1, databuf, sizeof(databuf), 0);
				if (nbBytesReceived <= 0)
				{
					CleanUp();
					return EXIT_FAILURE;
				}

				if (enablespy) 
				{
					if (charmode) 
					{
						for (i = 0; i < nbBytesReceived; i++)
						{
							if (fputc((int)databuf[i], fspy) == EOF)
							{
								printf("fputc() failed.\n");
							}
						}
					}
					else
					{
						for (i = 0; i < nbBytesReceived; i++)
						{
							if (fprintf(fspy, "%02x ", (unsigned int)((unsigned char)databuf[i])) <= 0)
							{
								printf("fprintf() failed.\n");
							}
						}
					}
					if (fflush(fspy) != EXIT_SUCCESS)
					{
						printf("fflush() failed.\n");
					}
				}

				if (sendall(sockcli, databuf, nbBytesReceived) != EXIT_SUCCESS)
				{
					CleanUp();
					return EXIT_FAILURE;
				}
				break;
			}
		}
	}

	if (!disablecli2) 
	{
		if (releasetcpcli(s2) != EXIT_SUCCESS)
		{
			releasetcpcli(s1);
			if (enablespy) fclose(fspy);
			return EXIT_FAILURE;
		}
	}
	if (releasetcpcli(s1) != EXIT_SUCCESS)
	{
		if (enablespy) fclose(fspy);
		return EXIT_FAILURE;
	}
	if (enablespy) 
	{
		if (fclose(fspy) != EXIT_SUCCESS) 
		{
			printf("fclose() failed.\n");
			return EXIT_FAILURE;
		}
	}

	return EXIT_SUCCESS;
}

int LoadConfig()
{
	FILE* file = NULL;
	char line[MAX_BUF_LEN];

	// Default values.
	readonly = 0;
	disablecli2 = 0;
	enablespy = 0;
	charmode = 0;
	timeout = 0;
	memset(srvport, 0, sizeof(srvport));
	sprintf(srvport, "27254");
	memset(cli1address, 0, sizeof(cli1address));
	sprintf(cli1address, "127.0.0.1");
	memset(cli1port, 0, sizeof(cli1port));
	sprintf(cli1port, "27255");
	memset(cli2address, 0, sizeof(cli2address));
	sprintf(cli2address, "127.0.0.1");
	memset(cli2port, 0, sizeof(cli2port));
	sprintf(cli2port, "27256");

	// Load data from a file.
	file = fopen("TCPSplit.txt", "r");
	if (file != NULL)
	{
		if (fgets3(file, line, sizeof(line)) == NULL) printf("Invalid configuration file.\n");
		if (sscanf(line, "%d", &readonly) != 1) printf("Invalid configuration file.\n");
		if (fgets3(file, line, sizeof(line)) == NULL) printf("Invalid configuration file.\n");
		if (sscanf(line, "%d", &disablecli2) != 1) printf("Invalid configuration file.\n");
		if (fgets3(file, line, sizeof(line)) == NULL) printf("Invalid configuration file.\n");
		if (sscanf(line, "%d", &enablespy) != 1) printf("Invalid configuration file.\n");
		if (fgets3(file, line, sizeof(line)) == NULL) printf("Invalid configuration file.\n");
		if (sscanf(line, "%d", &charmode) != 1) printf("Invalid configuration file.\n");
		if (fgets3(file, line, sizeof(line)) == NULL) printf("Invalid configuration file.\n");
		if (sscanf(line, "%d", &timeout) != 1) printf("Invalid configuration file.\n");
		if (fgets3(file, line, sizeof(line)) == NULL) printf("Invalid configuration file.\n");
		if (sscanf(line, "%255s", srvport) != 1) printf("Invalid configuration file.\n");
		if (fgets3(file, line, sizeof(line)) == NULL) printf("Invalid configuration file.\n");
		if (sscanf(line, "%255s", cli1address) != 1) printf("Invalid configuration file.\n");
		if (fgets3(file, line, sizeof(line)) == NULL) printf("Invalid configuration file.\n");
		if (sscanf(line, "%255s", cli1port) != 1) printf("Invalid configuration file.\n");
		if (fgets3(file, line, sizeof(line)) == NULL) printf("Invalid configuration file.\n");
		if (sscanf(line, "%255s", cli2address) != 1) printf("Invalid configuration file.\n");
		if (fgets3(file, line, sizeof(line)) == NULL) printf("Invalid configuration file.\n");
		if (sscanf(line, "%255s", cli2port) != 1) printf("Invalid configuration file.\n");

		//if (fgets3(file, line, sizeof(line)) == NULL) printf("Invalid configuration file.\n");
		//if (sscanf(line, "%lf", &val) != 1) printf("Invalid configuration file.\n");

		if (fclose(file) != EXIT_SUCCESS) printf("fclose() failed.\n");
	}
	else
	{
		printf("Configuration file not found.\n");
	}

	return EXIT_SUCCESS;
}

int main()
{
	LoadConfig();

	if (LaunchSingleCliTCPSrv(srvport, handlecli, NULL) != EXIT_SUCCESS)
	{
#ifdef _DEBUG
		fprintf(stdout, "Press ENTER to continue . . . ");
		(void)getchar();
#endif // _DEBUG
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
